# -*- coding: utf-8 -*-
## controller.py
## © Claus Albowski

## Importe:
import platform
versionsNummer=platform.python_version()[0]
if versionsNummer=='2':
    from Tkinter import *
    import tkFileDialog as fileDialog
else:
    from tkinter import *
    import tkinter.filedialog as fileDialog

from graphenprojekt.graphenModell import *
from graphenprojekt.fensterElemente import *

class Controller(object):
    
    derController=None
    
    @staticmethod
    def gibController(app):
        if Controller.derController==None: Controller.derController=Controller(app)
        return Controller.derController
    
    def __init__(self, app):
        '''Singleton - aber nicht erzwungen
        verwaltet die Verbindung zwischen den Objekten
        (Knoten und Kanten) und ihren CanvasElementen'''
        self.__knoten=dict()   # {name:(knoten, element)}
        self.__kanten=dict()  # {name:(kante, element)}
        
    def loescheAlle(self,zf):
        CanvasElement.loescheAlle(zf)
        kanten,alleKnoten=[],[]
        for name in self.__kanten: kanten.append(name)
        for name in kanten: self.loescheInKanten(name)
        for name in self.__knoten: alleKnoten.append(name)
        for name in alleKnoten: self.loescheInKnoten(name)
        
    def istLeer(self):
        return (len(self.__knoten)==0 and len(self.__kanten)==0)
    
    def gibKnotenUndElement(self,name):
        '''gibt zum Namen das Paar aus Knoten und Element zurueck'''
        return self.__knoten.get(name)

    def gibDatenMit(self, zf):
        '''haengt die zusaetzlichen Attribute an'''
        ## die Richtung war urprünglich nicht in den geholten Modelldaten
        ## enthalten (warum auch immer)
        alleDaten=Graph.gibGraphen().gibDaten()
        datenMit=[]
        for daten in alleDaten:
            if (daten[0]=='Knoten'):
                kn,ke=self.__knoten.get(daten[1][0])
                datenMit.append(('Knoten',
                                 daten[1]+
                                 (ke.gibFarbe(),
                                  ke.gibSichtbar())))
            elif (daten[0]=='Kante'):
                # BeschriftungsTyp,versatz,Farbe,richtung,d_PosText
                name=daten[1][0]+'-'+daten[1][1]
                ka,ke=self.__kanten.get(name)
                datenMit.append(('Kante',
                                 (daten[1][0],
                                  daten[1][1],
                                  ke.gibText(),
                                  ke.gibBeschriftungsTyp(),
                                  ke.gibVersatz(),
                                  ke.gibFarbe(),
                                  ke.gibRichtung(),
                                  ke.gibVerschiebungPositionText())))
            else:
                datenMit.append(daten)
        return datenMit
    
    def erzeugeGraphenAusDaten(self, zf, alleDaten):
        '''extrahiert die Werte und erzeugt die Elemente'''
        alle=eval(alleDaten) # String in Liste
        for daten in alle:
            if daten[0]=='Knoten':
                name=daten[1][0]
                pos=daten[1][1:3]
                if len(daten[1])>=4: farbe=daten[1][3]
                else: farbe=''
                if len(daten[1])==5: sichtbar=daten[1][4]
                else: sichtbar=True
                self.neuerKnoten(zf, name, pos, farbe, sichtbar)
            elif daten[0]=='Kante':
                if len(daten[1])==3:  ## ohne Darstellungs-Attribute 
                    von,auf,text = daten[1]
                    ka=self.neueKante(zf, von, auf, text)
                elif len(daten[1])==7:  ## ohne Textverschiebung
                    von,auf,text,beschriftungsTyp,versatz,farbe,richtung = daten[1]
                    if beschriftungsTyp=='Name': text=von+'-'+auf # sollte unnötig sein
                    ka=self.neueKante(zf, von,auf,text,beschriftungsTyp,versatz,farbe,richtung)
                else:  ## mit Textverschiebung
                    von,auf,text,beschriftungsTyp,versatz,farbe,richtung,dText=daten[1]
                    if beschriftungsTyp=='Name':
                        text=von+'-'+auf
                    ka=self.neueKante(zf, von,auf,text,beschriftungsTyp,versatz,farbe,richtung,dText)


    ## Farben
    def setzeFarbe(self, zf, name, neueFarbe):
        '''Knoten werden gefuellt'''
        if name in self.__knoten:
            kn,ke=self.__knoten.get(name)
            ke.setzeFarbe(zf, neueFarbe)
        elif name in self.__kanten:
            ka,ke=self.__kanten.get(name)
            ke.setzeFarbe(zf, neueFarbe)
        else:
            return 'Name unzulaessig'

    ## Knoten
    def neuerKnoten(self, zf, name, pos, farbe='', sichtbar=True):
        knoten=Graph.gibGraphen().neuerKnoten(name, pos)
        if knoten==None: return None
        element=KnotenElement(zf, name, pos, farbe, sichtbar)
        self.__knoten.update({name:(knoten, element)})
        return knoten

    def namenAnzeigen(self, zf, name, sichtbar=True):
        if not(name in self.__knoten.keys()): return None
        element=CanvasElement.nameUndElement[name]
        element.namenAnzeigen(zf, sichtbar)

    def alleKnotenNamenAnzeigen(self, zf, sichtbar=True):
        for name in CanvasElement.nameUndElement:
            element=CanvasElement.nameUndElement[name]
            if element.__class__==KnotenElement:
                element.namenAnzeigen(zf, sichtbar)

    def gibKnotenPosition(self, name):
        if not(name in self.__knoten.keys()): return None
        kn,ke=self.__knoten.get(name)
        return kn.gibDaten()[1:3]
    
    def loescheInKnoten(self, name):
        '''loescht nur im Knotendictionary'''
        if not(name in self.__knoten.keys()): return None
        return self.__knoten.pop(name)
    
    def loescheKnotenUndKanten(self, zf, knotenName):
        '''Loescht auch die zugehoerigen Kanten'''
        if not(knotenName in self.__knoten.keys()): return
        kantenNamen=[]
        for kantenName in self.__kanten:
            if knotenName in kantenName:
                kantenNamen.append(kantenName)
        for kantenName in kantenNamen:
            self.loescheInKanten(kantenName)
            CanvasElement.loescheElement(zf,kantenName)
        CanvasElement.loescheElement(zf, knotenName)
        Graph.gibGraphen().loescheKnoten(knotenName)
        self.loescheInKnoten(knotenName)
        
    def verschiebeKnotenUm(self, zf, name, dx, dy):
        kn,ke=self.gibKnotenUndElement(name)
        n,x,y=kn.gibDaten()
        kn.verschiebe(x+dx,y+dy)
        Graph.gibGraphen().knotenVerschoben(name) # aktualisiert Kanten
        farbe=ke.gibFarbe()
        CanvasElement.loescheElement(zf, name)
        self.loescheInKnoten(name) ### aus V2022-11-16
        element=KnotenElement(zf,name, (x+dx,y+dy),farbe)
        self.__knoten.update({name:(kn, element)})
        self.aendereKantenZu(zf, name, name)

    def verschiebeAlleKnotenUm(self, zf, dx, dy):
        alle=[]
        for name in self.__knoten:
            alle.append(name)
        for name in alle:
            self.verschiebeKnotenUm(zf, name, dx, dy)
        
    def aendereKnotenNamen(self, zf, alterName, neuerName):
        if alterName==neuerName: return
        if '-' in neuerName: return
        kn,ke=self.gibKnotenUndElement(alterName)
        Graph.gibGraphen().aendereKnotenNamen(alterName, neuerName)
        # aktualisiert in Graph Knoten und Kanten
        farbe=ke.gibFarbe()
        CanvasElement.loescheElement(zf, alterName)
        self.loescheInKnoten(alterName)
        element=KnotenElement(zf, neuerName, kn.gibDaten()[1:3], farbe)
        self.__knoten.update({neuerName:(kn, element)})
        self.aendereKantenZu(zf, alterName, neuerName)

    ## Kanten        
    def neueKante(self, zf, von, auf, text=' ', beschriftungsTyp=None,
                  versatz=0.0, farbe='',
                  richtung=None, verschiebungTextposition=(0,0)):
        '''erzeugt die Kanten im Modell (text fuer die Bewertung)
        und bei den Fensterelementen (text fuer die Beschriftung
        nach den Vorgaben des Beschriftungstyps)
        und aktualisiert das Dictionary'''
        kante=Graph.gibGraphen().neueKante(von, auf, text, richtung)
        if kante==None: return None
        name=von+'-'+auf
        if name in self.__kanten.keys(): return False
        element=KantenElement(zf, von, auf, text, beschriftungsTyp, versatz, farbe, richtung)
        self.__kanten.update({name:(kante, element)})
        if verschiebungTextposition!=(0,0):
            element.korrigiereTextPosition(zf, verschiebungTextposition[0], verschiebungTextposition[1])
        return kante

    def AlleKantenNeu(self,zf, alleKnotenNamen):
        for i in range(len(alleKnotenNamen)):
            for j in range(i+1,len(alleKnotenNamen)):
                self.neueKante(zf, alleKnotenNamen[i], alleKnotenNamen[j])

    def neueKantenBeschriftung(self, zf, name, beschriftungsTyp, text=' '):
        '''ändert den Typ der Beschriftung'''
        ka,ke=self.__kanten.get(name)
        ke.neuerBeschriftungsTyp(zf, beschriftungsTyp, text)
        ka.setzeBewertung(ke.gibText())
    
    def korrigiereTextPosition(self, zf, name, dx, dy):
        ka,ke=self.__kanten.get(name)
        ke.korrigiereTextPosition(zf,dx,dy)
        
    def loescheInKanten(self, name):
        '''loescht nur im Dictionary'''
        if not(name in self.__kanten.keys()): return None
        return self.__kanten.pop(name)

    def loescheKante(self, zf, name):
        '''loescht sowohl im Modell auch in der Darstellung'''
        if not(name in self.__kanten.keys()): return None
        ka,ke=self.__kanten.get(name)
        Graph.gibGraphen().loesche(ka)
        zf.delete(name)
        self.loescheInKanten(name)
    
    def gibKanteUndElement(self,name):
        ka,ke=self.__kanten.get(name)
        return name,ka,ke
    
    def gibAlleKantenZu(self, knotenName):
        passende=[]
        for name in self.__kanten:
            von,strich,auf=name.partition('-')
            if knotenName in (von,auf):
                passende.append(self.gibKanteUndElement(name))
        return passende

    def aendereKantenZu(self, zf, alt, neu):
        '''wird aufgerufen nach einer Aenderung eines Knotennamen
        oder Knotenposition'''
        alle=self.gibAlleKantenZu(alt)
        for dat in alle:
            von,strich,auf=dat[0].partition('-')
            text=dat[2].gibText()
            ka,ke=self.__kanten.get(dat[0])
            von,auf,text,be,ve,fa,ri,dxt,dyt=self.holeKantenElementDaten(ke)
            zf.delete(dat[0])
            self.loescheInKanten(dat[0]) ## wieder aufgenommen aus V2022-11-16
            if alt==neu:
                element=KantenElement(zf, von, auf, text,be,ve,fa,ri,(dxt,dyt))
            elif von==alt:
                element=KantenElement(zf,neu, auf, text,be,ve,fa,ri,(dxt,dyt))
            else:
                element=KantenElement(zf, von, neu, text,be,ve,fa,ri,(dxt,dyt))
            self.__kanten.update({element.gibTag():(ka,element)})

    def holeKantenElementDaten(self, ke):
        von,auf=ke.gibVon(),ke.gibAuf()
        text,be=ke.gibText(),ke.gibBeschriftungsTyp()
        ve,fa,ri=ke.gibVersatz(),ke.gibFarbe(),ke.gibRichtung()
        dxt,dyt=ke.gibVerschiebungPositionText()
        return von,auf,text,be,ve,fa,ri,dxt,dyt

    def aendereVersatz(self, zf, tag, versatz):
        ka,ke=self.__kanten.get(tag)
        von,auf,text,be,ve,fa,ri,dxt,dyt=self.holeKantenElementDaten(ke)
        zf.delete(ke.gibTag())
        self.kantenElementAktualisieren(zf, tag, ka, von, auf, text, be,
                                        versatz, # geaendert
                                        fa, ri, dxt, dyt)
        
    def aendereBogenOrientierung(self, zf, tag):
        ka,ke=self.__kanten.get(tag)
        von,auf,text,be,ve,fa,ri,dxt,dyt=self.holeKantenElementDaten(ke)
        zf.delete(ke.gibTag())
        self.kantenElementAktualisieren(zf, tag, ka, von, auf, text, be, -ve, # geaendert
                                        fa, ri, dxt, dyt)

    def aendereRichtung(self, zf, tag):
        ka,ke=self.__kanten.get(tag)
        von,auf,text,be,ve,fa,ri,dxt,dyt=self.holeKantenElementDaten(ke)
        zf.delete(ke.gibTag())
        richtung=ri
        if richtung=='first': richtung='last'
        elif richtung=='last': richtung='None'
        else: richtung='first'
        ka.setzeRichtung(richtung)
        self.kantenElementAktualisieren(zf, tag, ka, von, auf, text, be, ve, fa, richtung)
        
    def kantenElementAktualisieren(self, zf, tag, kante, von, auf, text , beschriftungsTyp, versatz, farbe, richtung, dxt=0, dyt=0):
        element=KantenElement(zf, von, auf, text, beschriftungsTyp, versatz, farbe, richtung,(dxt,dyt))
        self.__kanten.update({tag:(kante,element)})

    def kantenElementKorrigieren(self, zf, tag):
        '''Fehler in der Darstellung korrigieren'''
        ka,ke=self.__kanten.get(tag)
        von,auf,text,be,ve,fa,ri,dxt,dyt=self.holeKantenElementDaten(ke)
        zf.delete(ke.gibTag())
        self.kantenElementAktualisieren(zf, tag, ka, von, auf, text, be, ve, fa, ri, dxt, dyt)
