# -*- coding: utf-8 -*-
## fensterElemente.py
## © Claus Albowski

import platform
versionsNummer=platform.python_version()[0]
if versionsNummer=='3': from tkinter import *
else: from Tkinter import *
import math

from abc import ABC,abstractmethod
class CanvasElement(ABC):
    '''Alle Hauptelemente werden im Dictionary "nameUndElement" abgespeichert.
    Das Loeschen erfolgt dann ueber das tag.'''    

    @staticmethod
    def objektZuTag(tag):
        return CanvasElement.nameUndElement.get(tag)
    
    @staticmethod
    def neuesElement(tag, element):
        return CanvasElement.nameUndElement.update({(tag,element)})

    @staticmethod
    def loescheElement(zf, tag):
        element=CanvasElement.objektZuTag(tag)
        CanvasElement.alleFigurenIDs.remove(element.ID)
        CanvasElement.nameUndElement.pop(tag)
        zf.delete(tag)

    @staticmethod
    def loescheAlle(zf):
        alleTags=[]
        for tag in CanvasElement.nameUndElement:
            alleTags.append(tag)
        for tag in alleTags:
            CanvasElement.loescheElement(zf,tag)

    @staticmethod
    def istFigurID(ID):
        return ID in CanvasElement.alleFigurenIDs
    
    @staticmethod
    def istFigurTag(tag):
        return tag in CanvasElement.nameUndElement.keys()

    fg,bg='black','white'
    @staticmethod
    def Setze_FG_BG(fg,bg):
        CanvasElement.fg=fg
        CanvasElement.bg=bg    

    nameUndElement=dict()
    alleFigurenIDs = []
    
    def __init__(self, zf, ID, pos, endPos, name=None, farbe=None):
        CanvasElement.neuesElement(name,self)
        self.zf=zf
        self.ID=ID
        CanvasElement.alleFigurenIDs.append(ID)
        self.textID=None
        self.name=name
        self.x,self.y=pos
        self.xe,self.ye=endPos
        self.farbe=''
        self.initialisiereFarbe(zf, farbe)

    def initialisiereFarbe(self, zf, farbe):
        if farbe!=None:
            self.farbe=farbe
        else:
            self.farbe=''
        self.setzeFarbe(zf, self.farbe)

    def gibID(self): return self.ID
    def gibTag(self): return self.name
    def gibText(self): return self.text
    def gibPosition(self): return self.x,self.y
    def gibEndPosition(self): return self.xe,self.ye
    def gibMitte(self, koordinaten):
        x,y,xe,ye=koordinaten
        return int((x+xe)/2),int((y+ye)/2)
    def gibFarbe(self): return self.farbe

    @abstractmethod
    def setzeFarbe(self, zf, neueFarbe):
        '''englische Namen eingeben
        Veraendert wird die Fuellfarbe, Standardattribut ist fill
        outline waere bei Oval usw fuer die Umrandung moeglich,
        ist bei line unzulaessig
        Wegen des anderen Zugriffs fg/bg wird die Methode in den
        beiden Elementklassen unterschiedlich implementiert'''
        return NotImplemented
            
    def aendereText(self):
        self.zf.itemconfigure(self.textID, text=self.text)
    def FG(self): return CanvasElement.fg
    def BG(self): return CanvasElement.bg

####  KnotenElement ----------------------------------------
class KnotenElement(CanvasElement):

    radius = 14

    def __init__(self, zf, name, pos=(10,10), farbe='', sichtbar=True):
        '''zeichenflaeche, eindeutigen Namen und Position uebergeben
        Namen von Knotenelementen, die mit _K beginnen, werden
        nicht angezeigt!'''
        x,y,xe,ye=self.bestimmeKoordinaten(pos)
        if farbe=='': bg=self.BG()
        else: bg=farbe
        self.ID=zf.create_oval(x,y, xe,ye, width=3, tag=name,
                               fill=farbe, outline=self.FG())
        CanvasElement.__init__(self, zf, self.ID, (x,y), (xe,ye), name, farbe)
        self.__nameSichtbar=sichtbar
        xm,ym=(x+xe)//2,(y+ye)//2
        if len(name)==1:
            xt,yt=xm,ym
        else:
            xt,yt=xm+len(name)*6+10,ym+10
        if self.__nameSichtbar: text=name
        else: text=''
        self.textID=zf.create_text(xt,yt, text=text, tag=name,
                                   fill=self.FG(),font="Sans 14 bold")

    def gibSichtbar(self):
        return self.__nameSichtbar
    
    def namenAnzeigen(self, zf, sichtbar=True):
        self.__nameSichtbar=sichtbar
        if sichtbar:
            zf.itemconfigure(self.textID,text=self.gibTag())
        else:
            zf.itemconfigure(self.textID,text='')
        
    def bestimmeKoordinaten(self, pos):
        x,y=pos[0]-KnotenElement.radius,pos[1]-KnotenElement.radius
        xe,ye=pos[0]+KnotenElement.radius, pos[1]+KnotenElement.radius
        return x,y,xe,ye

    def setzeFarbe(self, zf, neueFarbe):
        '''s.o.'''
        self.farbe=neueFarbe
        if self.farbe=='':
            zf.itemconfig(self.gibID(), fill=self.BG())
        else:
            zf.itemconfig(self.gibID(), fill=neueFarbe)

####  KantenElement ----------------------------------------
class KantenElement(CanvasElement):

    def __init__(self, zf, von, auf, text=' ', beschriftungsTyp=' ',
                 versatz=0.0, farbe='',
                 richtung=NONE,dText=(0,0)):
        '''zeichenflaeche und eindeutige Namen von vorhandenen Knoten uebergeben
        "text" ohne oder externe Kantenbewertung,
        None fuer Berechnung des Abstands
        Richtung NONE, FIRST oder LAST'''
        self.von,self.auf,self.text,=von,auf,text
        self.beschriftungsTyp=beschriftungsTyp # Typ der Beschriftung
        self.dPosText=(0,0)
        if richtung=='first': richtung=FIRST
        elif richtung=='last': richtung=LAST
        else: richtung=NONE
        self.richtung=richtung
        self.farbe=farbe
        name=von+'-'+auf
        ## Mitte des Knoten "von" bestimmen
        knoten=CanvasElement.objektZuTag(von)
        x1,y1=knoten.gibPosition()
        x2,y2=knoten.gibEndPosition()
        self.x,self.y=(x1+x2)/2,(y1+y2)/2
        ## Mitte des Knoten "auf" bestimmen
        knoten=CanvasElement.objektZuTag(auf)
        x1,y1=knoten.gibPosition()
        x2,y2=knoten.gibEndPosition()
        self.xe,self.ye=(x1+x2)/2,(y1+y2)/2
        ## Enden der Linie bestimmen
        l=self.gibAbstand()
        x=self.x+(self.xe-self.x)*KnotenElement.radius/l
        y=self.y+(self.ye-self.y)*KnotenElement.radius/l
        xe=self.xe+(self.x-self.xe)*KnotenElement.radius/l
        ye=self.ye+(self.y-self.ye)*KnotenElement.radius/l
        # den Zwischenpunkt des Bogens orthogonal zur Mitte
        self.ve=versatz
        xz,yz=self.gibBogenMitte()
        if farbe=='': fg=self.FG()
        else: fg=farbe
        ID=zf.create_line(x,y,xz,yz,xe,ye,
                          arrow=self.richtung, arrowshape=(16,20,6),
                          smooth=True, width=3, tag=name)#, fill=fg)
        CanvasElement.__init__(self, zf, ID, (x,y), (xe,ye), name, farbe)
        zf.itemconfigure(ID,fill=fg)
        xt,yt=self.gibTextPosition()
        dx,dy=dText
        ##print(xz,yz,' / ',xt,yt)
        self.textID=zf.create_text(xt+dx,yt+dy, text=self.text, tag=name,
                                   fill=self.FG(), font="Sans 12 bold")
        self.neuerBeschriftungsTyp(zf, beschriftungsTyp, text)

    def gibVon(self): return self.von
    def gibAuf(self): return self.auf
    def gibBeschriftungsTyp(self): return self.beschriftungsTyp
    def gibVersatz(self): return self.ve
    def gibAbstand(self):
        dx,dy=self.xe-self.x,self.ye-self.y
        return math.sqrt(dx*dx+dy*dy)
    def gibRichtung(self): return self.richtung
    def gibVerschiebungPositionText(self): return self.dPosText

    def setzeFarbe(self, zf, neueFarbe):
        '''s.o.'''
        self.farbe=neueFarbe
        if self.farbe=='':
            zf.itemconfig(self.gibID(), fill=self.FG())
        else:
            zf.itemconfig(self.gibID(), fill=neueFarbe)

    def neuerBeschriftungsTyp(self, zf, neueVorgabe, text=' '):
        '''aendert den Beschriftungstyp'''
        ##print(zf, neueVorgabe, text)
        if neueVorgabe=='Abstand':
            self.beschriftungsTyp='Abstand'
            self.text=str(int(round(self.gibAbstand())))
        elif neueVorgabe=='Name':
            self.beschriftungsTyp='Name'
            self.text=self.gibVon()+'-'+self.gibAuf()
        elif neueVorgabe==' ':
            self.beschriftungsTyp=' '
            self.text=' '
        elif neueVorgabe=='extern':
            self.beschriftungsTyp=neueVorgabe
            self.text=text
        else: # sollte nicht passieren
            self.beschriftungsTyp=' '
            self.text=' '
        self.aendereText()

    def korrigiereTextPosition(self,zf,dx,dy):
        '''verschiebt den Text auf die uebergebenen Differenzwerte
        gegenüber der Normalposition; bei Uebergabe von (0,0)
        wird auf den Ausganswert zurueck gesetzt'''
        altX,altY=self.dPosText
        self.dPosText=dx,dy
        zf.move(self.textID, dx-altX, dy-altY)

    def gibBogenMitte(self):
        x,y,xe,ye=self.x,self.y,self.xe,self.ye
        xm , ym = self.gibMitte((self.x,self.y,self.xe,self.ye))
        xz , yz = xm+1+int(self.ve*(ye-y)), ym-1-int(self.ve*(xe-x))
        return xz,yz

    def gibTextPosition(self):
        x,y,xe,ye=self.x,self.y,self.xe,self.ye
        xm , ym = self.gibMitte((self.x,self.y,self.xe,self.ye))
        xz , yz = xm+1+int(0.6*self.ve*(ye-y)), ym-1-int(0.6*self.ve*(xe-x))
        if self.ve==0: return xz+self.dPosText[0]-10, yz+self.dPosText[1]+10
        return xz+self.dPosText[0], yz+self.dPosText[1]
    
