#!/usr/bin/env python
# -*- coding: utf-8 -*-
# graphenModell.py
# © Claus Albowski
from graphenprojekt.dialoge import * # eigene
import math

#### Klasse fuer beide Elemente ############################
class GraphenElement(object):
    '''Oberklasse fuer Knoten und Kanten'''    

    def __init__(self, name):
        '''fuer die Graphen werden die Knoten und Kanten
        und deren Namen gespeichert
        Name einer Kante zB: "A-B"'''
        self.name=name

    def gibNamen(self):
        try:
            return self.name
        except AttributeError as e:
            self.Meldung("Name ist nicht richtig initialisiert:\n"
                         +"SetzeNamen() mit String notwendig!")    

    def aendereNamen(self, neuerName):
        self.name=neuerName

#### Klasse für Knoten ====================================
class Knoten(GraphenElement):

    def __init__(self, name, x=10, y=10):
        GraphenElement.__init__(self, name)
        self.x , self.y = x , y

    def gibDaten(self):
        return self.gibNamen(), self.x , self.y
        
    def verschiebe(self, aufX, aufY):
        self.x , self.y = aufX, aufY

#### Klasse für Kanten ====================================
class Kante(GraphenElement):

    def __init__(self, vonKnoten, nachKnoten, bewertung=None, richtung=None):
        name=vonKnoten.gibNamen()+'-'+nachKnoten.gibNamen()
        GraphenElement.__init__(self, name) # setzt auch das Attribut
        self.vonKnoten = vonKnoten
        self.nachKnoten = nachKnoten
        self.richtung=richtung
        self.setzeBewertung(bewertung)

    def gibDaten(self):
        return self.vonKnoten.gibNamen(),\
               self.nachKnoten.gibNamen(),\
               self.gibBewertung(),\
               self.gibRichtung()
               
    def gibBewertung(self):
        if self.bewertung=='Abstand': return self.gibAbstand()
        else: return self.bewertung
    def gibVon(self): return self.gibDieKnoten()[0]
    def gibAuf(self): return self.gibDieKnoten()[1]
    def gibDieKnoten(self):
        '''gibt das Knotenpaar zurueck'''
        return self.vonKnoten,self.nachKnoten
    def gibRichtung(self): return self.richtung

    def setzeBewertung(self, neueBewertung=None):
        self.bewertung = neueBewertung

    def setzeRichtung(self, neueRichtung=None):
        self.richtung=neueRichtung
        
    def knotenNameGeaendert(self, alterName,neuerName):
        '''aendert den Namen der Kante, wenn sie betroffen ist'''
        if self.gibVon().gibNamen()==alterName:
            self.aendereNamen(neuerName+'-'+self.nachKnoten.gibNamen())
        elif self.gibAuf().gibNamen()==alterName:
            self.aendereNamen(self.vonKnoten.gibNamen()+'-'+neuerName)        
        return True

    def knotenVerschoben(self):
        '''notwendig, wenn die Bewertung aus dem Abstand besteht.'''
        if self.gibBewertung()!='Abstand': return
        self.setzeBewertung('Abstand')

    def gibAbstand(self):
        dx=self.nachKnoten.gibDaten()[1]-self.vonKnoten.gibDaten()[1]
        dy=self.nachKnoten.gibDaten()[2]-self.vonKnoten.gibDaten()[2]
        return math.sqrt(dx*dx+dy*dy)

#### Klasse für Graphen ===================================
class Graph(object):
    '''Modell-Komponente fuer die Graphen-Anwendung.'''

    derGraph = None

    @staticmethod
    def gibGraphen(name="EinGraph"):
        '''Graph ist Singleton'''
        if Graph.derGraph==None:
            Graph.derGraph = Graph(name)
        return Graph.derGraph
    
    #### Konstruktor von Graph
    def __init__(self, graphenName="Graph",view=None):
        self.__graphenName=graphenName
        self.__view=view
        self.__graphenDaten=[] # fuer die Ausgabe
        self.knoten = []
        self.kanten = []

    ## Methoden fuer Graph selbst
    def gibGraphenNamen(self):
        return self.__graphenName

    def setzeGraphenNamen(self, name):
        assert name.__class__==str, "String eingeben!"
        self.__graphenName=name
    
    def loescheGraphen(self):
        '''leert die beiden Listen und den Namen'''
        self.__graphenName=''
        self.knoten = []
        self.kanten = []

    def Meldung(self, text):
        Graph.gibGraphen().Meldung(text)

    ## Methoden fuer beide Graphenelemente
    def gibAlleNamen(self):
        namen=[]
        for element in self.knoten+self.kanten:
            namen.append(element.gibNamen())
        return namen

    def gibElementZuNamen(self, name):
        for element in (self.knoten + self.kanten):
            if name==element.gibNamen():
                return element

    def pruefeNamen(self, name):
        if name in self.gibAlleNamen():
            self.Meldung("Name existiert bereits!")
            return False
        try:
            assert name.__class__== str,\
                   "Der Name muss den Typ String haben."
            return True
        except AssertionError as e:
            self.Meldung(str(e))
            return False
    
    def loesche(self, element):
        if element in self.knoten:
             self.knoten.remove(element)
        elif element in self.kanten:
            self.kanten.remove(element)

    ## Methoden fuer Knoten
    def gibAlleKnoten(self):
        return self.knoten

    def gibAlleKnotenNamen(self):
        knotenNamen=[]
        for knoten in self.gibAlleKnoten():
            knotenNamen.append(knoten.gibNamen())
        return knotenNamen

    def gibKnotenZuNamen(self, name):
        element=self.gibElementZuNamen(name)
        if element in self.knoten: return element
        return None

    def neuerKnoten(self, name, pos=(10,10)):
        if not(self.pruefeNamen(name)): return None
        x,y=pos
        derKnoten = Knoten(name,x,y)
        self.knoten.append(derKnoten)
        return derKnoten
    
    def aendereKnotenNamen(self, alterName, neuerName):
        if not (alterName in self.gibAlleNamen()): return False # alterName existiert nicht
        einKnoten=self.gibElementZuNamen(alterName)
        if not (einKnoten in self.knoten): return False # ist kein Knoten
        for kante in self.kanten: # erst die Kantennamen aendern !
            kante.knotenNameGeaendert(alterName,neuerName)
        einKnoten.aendereNamen(neuerName)

    def loescheKnoten(self, name):
        '''entfernt das Knotenobjekt zum Namen
        und entfernt die zugehoerigen Kanten'''
        derKnoten=self.gibKnotenZuNamen(name)
        if derKnoten==None:
            self.Meldung("Knoten "+name+" existiert nicht!")
            return False
        ## Auch alle Kanten loeschen, die diesen Knoten betreffen.
        for kante in self.gibAlleKanten():
            if (name==kante.gibVon()) or (name==kante.gibAuf()):
                self.loescheKante(kante.gibVon(),kante.gibAuf())
        self.knoten.remove(derKnoten)
        return True
        
    ## Methoden fuer Kanten
    def gibKante(self, vonName, aufName):
        for kante in self.gibAlleKanten():
            if kante.gibVon()==vonName:
                if kante.gibAuf()==aufName:
                    return kante
            if kante.gibAuf()==vonName:
                if kante.gibVon()==aufName:
                    return kante
        return None
                
    def gibAlleKanten(self):
        return self.kanten

    def gibAlleKantenNamen(self):
        kantenNamen=[]
        for kanten in self.gibAlleKanten():
            kantenNamen.append(kanten.gibNamen())
        return kantenNamen

    def hatKante(self, vonName, aufName):
        return self.gibKante(vonName, aufName)!=None
    
    def neueKante(self, vonName, nachName, bewertung=None, richtung=None):
        if self.hatKante(vonName, nachName): return None # keine Mehrfachkanten
        if not(self.gibElementZuNamen(vonName) in self.knoten):
            return None # Knoten muss existieren
        if not(self.gibElementZuNamen(nachName) in self.knoten):
            return None # Knoten muss existieren
        dieKante = Kante(self.gibElementZuNamen(vonName),
                         self.gibElementZuNamen(nachName),
                         bewertung,
                         richtung)
        self.kanten.append(dieKante)
        return dieKante
        
    def loescheKante(self, vonName, aufName):
        kante=self.gibKante(vonName, aufName)
        if kante==None: return False # existiert nicht
        self.kanten.remove(kante)

    def knotenVerschoben(self, name):
        for kante in self.gibAlleKanten():
            if (kante.gibVon().gibNamen()==name
                or kante.gibAuf().gibNamen()==name):
                kante.knotenVerschoben()

#### Abschnitt für die Modelldefinition -----------------------------
    def gibDaten(self):
        self.sichereDaten() # Korrektur gegenueber V2022-11-16
        return self.__graphenDaten

    def zeigeDaten(self):
        '''Modelldaten zeilenweise lesbar''' ## Ausgabe unvollstaendig !!!!
        daten=self.__graphenDaten
        knoten,kanten=[],[]
        zeilen=' Daten\n\nKnoten:\n'
        for k in daten:
            if k[0]=='Knoten': knoten.append(k)
            elif k[0]=='Kante': kanten.append(k)
        for k in knoten: zeilen+=str(k[1])+'\n'
        zeilen+='\nKanten:\n'
        for k in kanten:
            richtung='--'
            if len(k[1])==4:
                if k[1][3]=='first': richtung='<='
                elif k[1][3]=='last': richtung='=>'
            zeilen+=str(k[1][0])+richtung+str(k[1][1])
            if (k[1][2]==' ') or (k[1][2]==str(k[1][0])+'-'+str(k[1][1])):
                zeilen+='\n'
            else:
                zeilen+=', Bewertung: '+str(k[1][2])+'\n'
        return zeilen
    
    def sichereDaten(self):
        self.__graphenDaten=[]
        if self.__graphenName==None:
            self.setzeGraphenNamen('Graph')
        for knoten in self.knoten:
            self.__graphenDaten.append(('Knoten',knoten.gibDaten()))
        for kante in self.kanten:
            self.__graphenDaten.append(('Kante',kante.gibDaten()))
        
    def erzeugeGraphenAusDaten(self, graphen_Daten):
        for daten in graphen_Daten:
            if daten[0]=='Knoten':
                self.neuerKnoten(daten[1][0],(daten[1][1],daten[1][2]))
            elif daten[0]=='Kante':
                self.neueKante(daten[1][0],daten[1][1],daten[1][2])
            else:
                self.Meldung('Fehler bei Graphendefinition: '+daten[0])

    def graphenNameAusDateiNamen(self, dateiname):
        # Pfad abspalten
        if '\\' in dateiname: # windows-Pfadangabe
            liste=dateiname.rsplit('\\')
        else:
            liste=dateiname.rsplit('/')
        liste.reverse()
        name=liste[0]
        if '.' in name:
            name=name.rsplit('.')[0]
        return name

#### Abschnitt für das Laden und Speichern von Modelldaten -----------------
    def speichere(self, dateiname='', graphen_Daten=[]):
        '''Das Modell allein enthaelt nicht alle Daten der Darstellung
        Daher koennen hier Daten einschliesslich der erweiterten
        Kanteneigenschaften gespeichert und gelesen werden.
        Im Modell selbst werden aber nur die dafuer wichtigen Daten verwendet.'''
        self.sichereDaten()
        if dateiname=='':
            if self.__graphenName==None: return
            dateiname='GraphenDaten/'+self.__graphenName+'.graph'
        try:
            f=open(dateiname,'w')
        except IOError as e:
            self.Meldung('Datei existiert nicht, keine Daten geschrieben')
            return None
        if graphen_Daten!=[]: f.write(str(graphen_Daten))
        else: f.write(str(self.__graphenDaten))
        f.close()

    def schreibeTextlich(self, dateiname=''):
        '''Textliche Speicherung des Modells
        Im Modell selbst werden aber nur die dafuer wichtigen Daten verwendet.'''
        self.sichereDaten()
        if dateiname=='':
            if self.__graphenName==None: return
            dateiname='DatenText/Daten_'+self.__graphenName+'.txt'
        try:
            f=open(dateiname,'w')
            text='Graph: '+self.__graphenName+'\n'
            for daten in self.zeigeDaten():
                text+='\n'+str(daten)
            f.write(text)
            f.close()
        except IOError as e:
            self.Meldung('Datei existiert nicht, keine Daten geschrieben')
            return None

    def lies(self, dateiname, datenAnApp=False):
        '''Verwendet nur die fuer das Modell selbst notwendigen Daten'''
        try:
            f=open(dateiname,'r')
        except IOError as e:
            self.Meldung('Datei existiert nicht, keine Daten gelesen')
            return None
        graphen_Daten_String=f.read()
        f.close()
        if datenAnApp: return graphen_Daten_String
        self.__graphenName=self.graphenNameAusDateiNamen(dateiname)
        self.__graphenDaten=eval(graphen_Daten_String)
        self.erzeugeGraphenAusDaten(self.__graphenDaten)
        return self.__graphenDaten

#### Zusaetzliches -----------------------------
    def ZeigeHilfe(self, n=1):
        if n==1:
            hinweise=("TODO"
                      )
        elif n==2:
            hinweise=("TODO.")
        return hinweise
    
    def SetzeView(self, view):
        self.__view=view
        
    def Meldung(self, text):
        if self.__view==None: print(text)
        else:
            MeldungDialog(self.__view,text)
            

## Tests
if __name__=='__main__':
    graph=Graph.gibGraphen()
##    graph.neuerKnoten("A",(100,100))
##    graph.neuerKnoten("B",(200,150))
##    graph.neuerKnoten("C",(150,250))
##    graph.neueKante("A","B",'Abstand')
##    graph.neueKante("A","C",33.3,'last')
##    graph.neueKante("B","C",None)
##    graph.sichereDaten()
##    print(graph.zeigeDaten())
##    graph.gibAlleNamen()
##    graph.aendereKnotenNamen('C','ccc')
##    graph.gibAlleNamen()
####    for knoten in graph.gibAlleKnoten():
####        graph.Meldung(knoten.gibDaten())
####    for kante in graph.gibAlleKanten():
####        graph.Meldung(kante.gibDaten())
##    graph.sichereDaten()
####    for daten in graph.zeigeDaten(): graph.Meldung(daten)
##    k=graph.gibKnotenZuNamen("B")
##    k.verschiebe(300,200)
##    graph.sichereDaten()
##    for daten in graph.zeigeDaten(): graph.Meldung(daten)
##    graph.Meldung(graph.gibAlleKnotenNamen())
##    graph.speichere("GraphenDaten/Test.graph",graph.gibDaten())
##    graph.schreibeTextlich()
##    graph.loescheGraphenDefinition()
    graph.lies('GraphenDaten/Test.graph')
    graph.sichereDaten()
    print(graph.zeigeDaten())
